import logging

from sqlalchemy import exc as sqla_exc
from sqlalchemy import create_engine as sqla_create_engine
from sqlalchemy import pool as sqla_pool
from sqlalchemy import orm

from anodyne import exceptions as _exceptions

on_engine_failure = None

logger = logging.getLogger(__name__)
server_engines = {}
backends = {}
default_server_engine = None
configured = False


class engine_types:
    postgres = "postgresql"
    postgres_psycopg2 = "postgresql+psycopg2"
    mysql = "mysql"
    pymysql = "mysql+pymysql"
    sqlite = "sqlite"

_engine_types = [
    engine_types.postgres,
    engine_types.postgres_psycopg2,
    engine_types.mysql,
    engine_types.pymysql,
    engine_types.sqlite
]


def configure(server_backends, default_engine=None, failure_callback=None):
    global on_engine_failure, backends, configured, default_server_engine
    if configured:
        raise _exceptions.ConfigurationException(
            "The connection layer has already been configured."
        )

    if backends is None:
        raise ValueError("server_backends cannot be None!")

    backends = server_backends
    on_engine_failure = failure_callback

    if default_engine is not None and default_engine in _engine_types:
        default_server_engine = default_server_engine

    configured = True


def create_engine(engine_type, connection_details, log_name_prefix=None):
    """
    Creates an engine based on the provided engine type and connection details.

    :param engine_type: a tincture.engines.engine_types string.
    :param connection_details: a dictionary containing connection details.
            Must include keys: user, passwd, host, port and db.
    :return: a dictionary of engine meta data, including the engine reference.
    """
    global configured
    if not configured:
        raise _exceptions.ConfigurationException(
            "Cannot create engines before configuration."
        )
    if engine_type == engine_types.sqlite:
        url = connection_details.get("url")
        if url is not None:
            connection_string = "%s:///%s" % (
                engine_type, connection_details.get("dbname")
            )
        else:
            connection_string = url
    else:
        connection_string = "%s://%s:%s@%s:%s/%s" % (
            engine_type,
            connection_details.get("username"),
            connection_details.get("password"),
            connection_details.get("host"),
            connection_details.get("port"),
            connection_details.get("dbname")
        )
    kwargs = {
        "poolclass": sqla_pool.QueuePool,
    }
    if log_name_prefix is not None:
        kwargs["logging_name"] = "%s.%s" % (__name__, log_name_prefix)

    engine = sqla_create_engine(connection_string, **kwargs)
    engine_data = {
        "failed": False,
        "engine_type": engine_type,
        "connection_details": connection_details,
        "engine": engine,
        "session_class": orm.sessionmaker(bind=engine)
    }
    return engine_data


def poke_engine(engine_meta_data, attempt):
    """
    "Poking" an engine. Meaning attempting a connection on said engine in order
    to revive it's health status (failed => False).

    This function is useful for periodic health checks on a given engine,
    as well as recovery of a dead engine. It will call the global callback
    function `failure_callback`, in the event that the connection fails.

    It is useful to configure this as something such as "gevent.spawn_later",
    using this function as the callback.

    :param engine_meta_data: the engine_meta_data to check.
    :param attempt: how many times we've already poked this engine.
    """
    global on_engine_failure, configured
    if not configured:
        raise _exceptions.ConfigurationException(
            "Cannot poke engine before configuration."
        )

    engine = engine_meta_data.get("engine")
    logger.info("Poking engine: %r. This is attempt: %d" % (
        repr(engine), (attempt + 1)
    ))
    try:
        conn = engine.connect()
        conn.close()
        engine_meta_data["failed"] = False
        logger.info("Engine revived.")
    except sqla_exc.OperationalError:
        logger.warning(
            "Poked %r, but it still appears to be down." % repr(engine)
        )
        if on_engine_failure is not None:
            on_engine_failure(engine_meta_data, attempt + 1)


def mark_failed(server_list_key, engine):
    """
    Marks a given engine as failed, If the failure callback is defined, it will
    be executed with the engine object.

    :param server_list_key: a db name mapping to connection engine meta data.
    :param engine: engine metadata that should have a reference to.
    """
    global server_engines, on_engine_failure, configured
    if not configured:
        raise _exceptions.ConfigurationException(
            "Cannot mark failures before configuration."
        )

    if engine is None:
        logger.error("Cannot mark %s engine meta as failed" % type(None))
        return

    if server_list_key is None:
        logger.error("Need a server in order to mark a failed engine.")

    engines = server_engines.get(server_list_key)

    if engines is None:
        logger.error("No engines configured for: %s." % server_list_key)
        return

    if engine not in engines:
        logger.error("Engine: %r does not belong to %r" % (
            engine.get("engine"), server_list_key
        ))

    engine["failed"] = True
    if on_engine_failure is not None:
        on_engine_failure(engine, 0)


def shuffle_engines(server_list_key):
    """
    Shuffles the engines in an engine list mapped by server_list_key.

    :param server_list_key: a db name mapping to connection engine meta data.
    """
    global server_engines, configured
    if not configured:
        raise _exceptions.ConfigurationException(
            "Cannot shuffle engines before configuration."
        )

    engine_list = server_engines.get(server_list_key, [])
    if len(engine_list) > 1:
        being_recycled = engine_list.pop(0)
        logger.info("Shuffling database engine: %s to %s" % (
            being_recycled,
            engine_list[0]
        ))
        engine_list.append(being_recycled)
        server_engines[server_list_key] = engine_list
    else:
        logger.warning(
            "Cannot shuffle engines with less than 2 servers defined."
        )


def get_engine(server_list_key, recycle=False, log_name=""):
    """
    Retrieves a sqlalchemy Engine object to use for database connections.

    :param server_list_key: a db name mapping to connection engine meta data.
    :param recycle: Whether we should recycle the engines for server_list_key.
    :return: engine meta data used to create a connection.
    """
    global server_engines, configured, backends, default_server_engine
    if not configured:
        raise _exceptions.ConfigurationException(
            "Cannot get engines before configuration."
        )
    database_engine = None
    if default_server_engine is not None:
        database_engine = default_server_engine

    server_info = backends.get(server_list_key)
    database_engine = server_info.get("engine_type", database_engine)
    engines = server_engines.get(server_list_key)

    if engines is None:
        server_list = server_info.get("servers")
        if server_list is None:
            server_list = []

        engine_list = []
        for index, engine_def in enumerate(server_list):
            one_based = index + 1
            num_prefix = "0%d" % one_based if index < 10 else one_based
            if log_name.startswith("-"):
                num_prefix = "%s-" % num_prefix

            log_prefix = "%s%s-%s" % (
                num_prefix,
                log_name,
                server_list_key
            )
            new_engine = create_engine(
                database_engine,
                engine_def,
                log_name_prefix=log_prefix
            )
            engine_list.append(new_engine)
        server_engines[server_list_key] = engine_list
    else:
        if recycle:
            shuffle_engines(server_list_key)

    engines = server_engines[server_list_key]
    for engine in engines:
        if not engine.get("failed", False):
            return engine
    return None


def clean_engines(server_list_key=None):
    """
    Disposes of all listed engines mapped to by server_list_key.
    If server_list_key is empty, we clear them all.

    :param server_list_key: a db name mapping to connection engine meta data.
    """
    global server_engines, configured
    if not configured:
        raise _exceptions.ConfigurationException(
            "Cannot clean engines before configuration."
        )
    if server_list_key is None:
        keys_to_clear = server_engines.keys()
    else:
        keys_to_clear = [server_list_key]
    for key in keys_to_clear:
        engine_list = server_engines.get(key)
        if engine_list is not None:
            for engine_data in engine_list:
                engine_data["engine"].dispose()
            del server_engines[key]
        else:
            # raise an exception?
            logger.error(
                "Tried to clear engines for keys: '%s', but key was not "
                "found." % keys_to_clear
            )



import contextlib
import logging

from anodyne import exceptions
from anodyne import engines

from sqlalchemy import exc as sqla_exc

logger = logging.getLogger(__name__)


def _get_engine(database, recycle=False):
    """
    Convenience function for retrieving a database engine from tincture.engines.

    :param database: The database to retrieve an engine for.
    :param recycle: Whether or not to direct the engines to be recycled.
    :return: a database engine object.
    """
    engine_data = engines.get_engine(database, recycle=recycle)
    if engine_data is None:
        err = "No engines available for '%s'" % database
        raise exceptions.NoValidEnginesException(err)
    if engine_data.get("engine") is None:
        err = "Engine Data found, but no engine was created. %s" % engine_data
        raise exceptions.EmptyEngineException(err)
    return engine_data


def get_connection(database):
    """
    Retrieves a database connection for the database reference provided.

    :param database: a name of database engine managed by tincture.
    :return: a connection.
    """
    engine_data = _get_engine(database)
    engine = engine_data.get("engine")
    try:
        conn = engine.connect()
    except sqla_exc.OperationalError, ex:
        logger.exception(
            "Engine failed with error: %r. Switching over." % repr(ex)
        )
        engines.mark_failed(database, engine_data)
        engine_data = _get_engine(database, recycle=True)
        engine = engine_data.get("engine")
        try:
            conn = engine.connect()
        except sqla_exc.OperationalError:
            engines.mark_failed(database, engine_data)
            logger.exception("Failed to connect to the database")
            raise
    return conn


@contextlib.contextmanager
def connection(database):
    """
    Context manager for a database connection.

    Used by the:
    with connection("Foo") as conn:
        conn.execute("...")
        ...
    paradigm.

    Begins a database transaction for the requester. Cleans up appropriately
    on error or successful transaction.

    :param database: a name of database engine managed by tincture.
    :yields: a database connection to the requester.
    """
    conn = get_connection(database)
    transaction = conn.begin()
    try:
        yield conn
    except sqla_exc.OperationalError:
        logger.exception("Failed to execute sql statement.")
        transaction.rollback()
        conn.close()
        raise
    except sqla_exc.DBAPIError:
        logger.exception("Failed to execute transaction.")
        transaction.rollback()
        conn.close()
        raise
    except Exception:
        logger.exception("Unknown exception occurred during transaction.")
        transaction.rollback()
        conn.close()
        raise
    else:
        transaction.commit()
        conn.close()


def get_orm_session(database):
    """
    Retrieves a sqlalchemy orm session for the database reference provided.

    :param database: a name of database engine managed by tincture.
    :return: a session.
    """
    engine_data = _get_engine(database)
    engine = engine_data.get("engine")
    try:
        engine.connect()
        session = engine_data.get("session_class")()
    except sqla_exc.OperationalError, ex:
        logger.exception(
            "Engine failed with error: %r. Switching over." % repr(ex)
        )
        engines.mark_failed(database, engine_data)
        engine_data = _get_engine(database, recycle=True)
        engine = engine_data.get("engine")
        try:
            engine.connect()
            session = engine_data.get("session_class")()
        except sqla_exc.OperationalError:
            engines.mark_failed(database, engine_data)
            logger.exception("Failed to connect to the database")
            raise
    return session


@contextlib.contextmanager
def orm_session(database):
    """
    Context manager for a database connection.

    Used by the:
    with orm_session("Foo") as session:
        session.add(object)
        ...
    paradigm.

    Begins a sqlalchemy orm session for the requester. Cleans up appropriately
    on error or successful transaction.

    :param database: a name of database engine managed by tincture.
    :yields: a database connection to the requester.
    """
    session = get_orm_session(database)
    connection = session.connection()
    transaction = connection.begin()
    try:
        yield session
    except sqla_exc.OperationalError:
        logger.exception("Failed to execute sql statement.")
        transaction.rollback()
        raise
    except sqla_exc.DBAPIError:
        logger.error("Failed to execute transaction.")
        transaction.rollback()
        raise
    else:
        session.commit()
    finally:
        session.close()
        connection.close()



import collections


_transfer_types = {}


def _new_transfer_obj(cls, obj_instance, **kwargs):
    """
    The purpose of this function is to serve as the __new__ method for a
    generated runtime class derived from a collections.namedtuple.

    :param cls: runtime class.
    :param obj_instance: instance of a sqlalchemy class.
    :param kwargs: keyword args to provide values for "extras"
    :return: an immutable data transfer object instance.
    """
    # The parent class invocation takes an expanded tuple of values that should
    # map 1-to-1 in position with the provided attributes.
    vals = ()
    defaults = getattr(cls, "_defaults", {})
    # Need to invoke the superclass tuple with values in order of the parent's
    # namedtuple specification.
    for key in getattr(cls, "_column_keys", []):
        # Default to None if we can't find the key anywhere.
        val = None
        if key in obj_instance.keys():
            # Try for the sqlalchemy class attribute first.
            val = obj_instance[key]
        elif key in getattr(cls, "_extras_keys", []):
            # Try for the "extras" next.
            if key in kwargs.keys():
                # If they gave it to us, use that value.
                val = kwargs.get(key)
            else:
                # Else, try a default.

                default_val = defaults.get(key)
                if default_val is not None:
                    val = default_val
        vals += (val,)

    # This is equivalent to calling "super(Foo, self).fn(*args, **kwargs)"
    # When the class Foo was defined as a class instead of generated.
    _invocation_type = cls.__mro__[0]
    return super(_invocation_type, cls).__new__(_invocation_type, *vals)


def _data_class(sqlalchemy_class, excludes=None, extras=None, name=None,
                mixin=None):
    """
    Generates a runtime class which others will instantiate on a sqlalchemy
    row proxy.

    :param sqlalchemy_class: a sqlalchemy table object.
    :param excludes: array for excluding columns from sqlalchemy_class.
    :param extras: a list (or dict of defaults) of extra attributes.
    :param name: an optional name parameter (mostly for debugging class names).
    :return: a class to instantiate.
    """
    global _transfer_types
    if name is None:
        name = "DataClass_%s" % sqlalchemy_class.name

    # Check if we've already created this class before.
    data_class = _transfer_types.get(name)
    if data_class is not None:
        return data_class

    # Base keys, pulled from the sqlalchemy class.
    _base_keys = [col.name for col in sqlalchemy_class.columns]
    _excludes = excludes or []
    _defaults = {}
    _extras_keys = []
    if extras is not None:
        if isinstance(extras, list):
            _extras_keys = extras
        elif isinstance(extras, dict):
            _extras_keys = extras.keys()
            _defaults = extras

    # remove keys from the set of base keys.
    _base_cols = set.difference(set(_base_keys), set(_excludes))
    _excluded = set.intersection(set(_base_keys), set(_excludes))
    _extras_keys = set.difference(set(_extras_keys), set(_base_cols))
    _extras_keys = set.difference(set(_extras_keys), set(_excludes))
    column_keys = set.union(set(_base_cols), set(_extras_keys))

    tuple_attributes = " ".join(column_keys)
    # The following lines create a generated runtime class.

    # The name of the parent class type will be "anodyne.data.<name>_tuple"
    type_name = "%s_tuple" % name
    data_class_type = collections.namedtuple(type_name, tuple_attributes)

    data_class_dict = dict(
        # How to generate a new instance.
        __new__=_new_transfer_obj,
        # class attributes for internal use and some inspection/debugging.
        _excluded=_excluded,
        _defaults=_defaults,
        _column_keys=column_keys,
        _extras_keys=_extras_keys,
        _table=sqlalchemy_class.fullname,
        # list'd class properties for inspection/debugging.
        extra_keys=lambda self: list(self._extras_keys),
        columns=lambda self: list(self._column_keys),
        excluded_columns=lambda self: list(self._excluded)
    )
    type_classes = (data_class_type,)
    if mixin:
        type_classes += (mixin,)

    data_class = type(
        str(name),  # The classtype name will be "anodyne.data.<name>"
        type_classes,  # Inherit from the generated transfer type.
        # This is what eventually updates __dict__ in the class/instance.
        data_class_dict
    )
    # Cache the class by its name, so we don't need to generate it again later.
    _transfer_types[name] = data_class
    return data_class


DataClass = _data_class



import sys

from sqlalchemy.ext import declarative
from anodyne import engines

ModelBase = declarative.declarative_base()


def setup(database):
    engine_data = engines.get_engine(database)
    ModelBase.metadata.create_all(engine_data.get("engine"))


def load_models(database, models_module):
    if models_module is not None:
        if models_module not in sys.modules.keys():
            __import__(models_module)
    engine_data = engines.get_engine(database)
    ModelBase.metadata.create_all(engine_data.get("engine"))



import logging

from sqlalchemy import exc as sqla_exc
from sqlalchemy import event as sqla_event
from sqlalchemy import pool as sqla_pool

logger = logging.getLogger(__name__)


@sqla_event.listens_for(sqla_pool.Pool, "checkout")
def disconnect_rebound(dbapi_connection, connection_record, connection_proxy):
    """
    Stolen largely from:
        http://docs.sqlalchemy.org/en/rel_0_8/core/pooling.html#disconnect-handling-pessimistic

    The purpose of this callback is to attempt health checks on a cursor
    before it is given back to an engine from a connection pool. The
    default behavior is to attempt a SELECT 1 cursor execution up to three
    times before re-raising or notifying the requester of a bad connection.

    :param dbapi_connection:
    :param connection_record:
    :param connection_proxy:
    :return:
    """
    cursor = dbapi_connection.cursor()
    try:
        cursor.execute("SELECT 1")
    except:
        # optional - dispose the whole pool
        # instead of invalidating one at a time
        # connection_proxy._pool.dispose()

        # raise DisconnectionError - pool will try
        # connecting again up to three times before raising.
        logger.exception(
            "Failed to establish connection for %r" % repr(dbapi_connection)
        )
        raise sqla_exc.DisconnectionError()
    cursor.close()


import logging

logger = logging.getLogger(__name__)


def save_instance(session, instance):
    """
    Save an instance of a model object.
    If we already have staged changes, commit them first and then save.
    """
    created = instance.id is None
    success = False
    try:
        session.merge(instance)
        session.commit()
        success = True
    except Exception:
        logger.exception("Exception while attempting to save an instance.")
        session.rollback()
        created = False

    if not success:
        logger.error("Failed to save instance! TODO PRINT ME SOME MORE INFO")

    return success, created


def query(session, model, **kwargs):
    try:
        if len(kwargs.keys()):
            return session.query(model).filter_by(**kwargs).all()
        else:
            return session.query(model).all()
    except:
        session.expunge()
        raise


def get(session, model, **kwargs):
    results = query(session, model, **kwargs)
    if len(results) > 1:
        raise Exception("Query for: '%s' returned more than 1 result!")
    if len(results):
        return results[0]
    return None


from setuptools import setup


VERSION = (0, 0, 4, "")

requirements = [
    "SQLAlchemy==0.9.7"
]

setup(
    name="anodyne",
    description="SQLAlchemy Database Utilities",
    packages=["anodyne"],
    version=".".join(filter(None, map(str, VERSION))),
    author="Alex Milstead",
    author_email="alex@amilstead.com",
    maintainer="Alex Milstead",
    maintainer_email="alex@amilstead.com",
    url="https://github.com/amilstead/anodyne",
    keywords=["sqlalchemy"],
    install_requires=requirements,
    package_dir={"anodyne": "anodyne"},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Topic :: Database ",
        "License :: OSI Approved :: GNU General Public License (GPL)"
    ],
)



class ConfigurationException(Exception):
    """
    Happens when someone attempts to use any core functions out of the `engines`
    module, but the system has yet to be configured.
    """


class EmptyEngineException(Exception):
    """
    Occurs when someone asks for a connection, and the engine being used
    to create that connection was not generated for some reason.
    """


class NoValidEnginesException(Exception):
    """
    Should be raised when connections attempt to retrieve engine metadata
    and that metadata is empty.
    """